#include <ArduinoOTA.h>

#include <ESPmDNS.h>

#include <LittleFS.h>
#include <configure.h>
#include <wifi.h>

#include <string.h>
#include <WiFi.h>
#include <WiFiUdp.h>

#include <LedControl_SW_SPI.h>

#include <ESP32-HUB75-MatrixPanel-I2S-DMA.h>

#define PANEL_RES_X 64      // Number of pixels wide of each INDIVIDUAL panel module. 
#define PANEL_RES_Y 32     // Number of pixels tall of each INDIVIDUAL panel module.
#define PANEL_CHAIN 1      // Total number of panels chained one to another

#include "font.h"

MatrixPanel_I2S_DMA *dma_display = nullptr;

constexpr const int pixelFloodPortAscii = 9810;
constexpr const int pixelFloodPortBin   = 9812;
constexpr const int binFramesPort       = 9811;

WiFiUDP UDPPixelFloodAscii;
WiFiUDP UDPPixelFloodBin;
WiFiUDP UDPBinFrames;

LedControl_SW_SPI lc;

uint32_t last_ts = 0;

bool ota_enabled = false;

bool progress_indicator(const int nr, const int mx, const std::string & which) {
	printf("%3.2f%%: %s\r\n", nr * 100. / mx, which.c_str());

	return true;
}

void setup_wifi() {
	set_hostname("RookScherm");

	enable_wifi_debug();

	scan_access_points_start();

	if (!LittleFS.begin())
		printf("LittleFS.begin() failed\r\n");

	configure_wifi cw;

	if (cw.is_configured() == false) {
retry:
		start_wifi("RookScherm");  // enable wifi with AP (empty string for no AP)

		cw.configure_aps();
	}
	else {
		start_wifi("");
	}

	// see what we can see
	printf("scan\r\n");
	scan_access_points_start();

	while(scan_access_points_wait() == false)
		delay(100);

	auto available_access_points = scan_access_points_get();

	// try to connect
	printf("connect\r\n");

	auto state = try_connect_init(cw.get_targets(), available_access_points, 300, progress_indicator);

	connect_status_t cs = CS_IDLE;

	for(;;) {
		cs = try_connect_tick(state);

		if (cs != CS_IDLE)
			break;

		delay(100);
	}

	// could not connect, restart esp
	// you could also re-run the portal
	if (cs == CS_FAILURE) {
		Serial.println(F("Failed to connect."));

		goto retry;

#if defined(ESP32)
		ESP.restart();
#else
		ESP.reset();
#endif
	}

	// connected!
	Serial.println(F("Connected!"));

	WiFi.setSleep(false);

	ArduinoOTA
		.onStart([]() {
				ota_enabled = true;

				dma_display->stopDMAoutput();

				String type;
				if (ArduinoOTA.getCommand() == U_FLASH)
					type = "sketch";
				else // U_SPIFFS
					type = "filesystem";

				// NOTE: if updating SPIFFS this would be the place to unmount SPIFFS using SPIFFS.end()
				LittleFS.end();
				Serial.println("Start updating " + type);
			})
	.onEnd([]() {
			Serial.println("\nEnd");
			})
	.onProgress([](unsigned int progress, unsigned int total) {
			Serial.printf("Progress: %u%%\r", (progress / (total / 100)));
			})
	.onError([](ota_error_t error) {
			Serial.printf("Error[%u]: ", error);
			if (error == OTA_AUTH_ERROR) Serial.println("Auth Failed");
			else if (error == OTA_BEGIN_ERROR) Serial.println("Begin Failed");
			else if (error == OTA_CONNECT_ERROR) Serial.println("Connect Failed");
			else if (error == OTA_RECEIVE_ERROR) Serial.println("Receive Failed");
			else if (error == OTA_END_ERROR) Serial.println("End Failed");
			});

	ArduinoOTA.begin();

	if (!MDNS.begin("RookScherm"))
	      Serial.println(F("Error setting up MDNS responder!"));
}

void printText(const char *const str) {
	const int len = strlen(str);

	int x_offset = 0;
	int y_offset = 0;

	for(int i=0; i<len; i++) {
		for(int y=0; y<8; y++) {
			for(int x=0; x<7; x++)
				dma_display->drawPixelRGB888(x_offset + x, y_offset + y, 0, font[str[i]][y][x], 0);
		}

		x_offset += 7;

		if (x_offset >= 64 - 6) {
			x_offset = 0;

			y_offset += 8;
			if (y_offset >= 32)
				return;
		}
	}
}

void setup() {
	Serial.begin(115200);

	Serial.setDebugOutput(true);

	Serial.println(F("Init"));

	// void begin(int dataPin, int clkPin, int csPin, int numDevices=1);
	// blauw CLK
	// wit   CS
	// paars DIN
	lc.begin(32, 18, 33, 8);

	for(int i=0; i<8; i++) {
		lc.shutdown(i, false);
		lc.setIntensity(i, 8);
		lc.clearDisplay(i);
	}

	// Module configuration
	HUB75_I2S_CFG mxconfig(
			PANEL_RES_X,   // module width
			PANEL_RES_Y,   // module height
			PANEL_CHAIN    // Chain length
			);

	mxconfig.gpio.e = -1;
	//mxconfig.clkphase = false;
	//mxconfig.driver = HUB75_I2S_CFG::FM6126A;

	// Display Setup
	dma_display = new MatrixPanel_I2S_DMA(mxconfig);
	dma_display->begin();
	dma_display->setBrightness8(180); //0-255
	dma_display->clearScreen();

	setup_wifi();

	WiFi.setSleep(false);

	UDPPixelFloodAscii.begin(pixelFloodPortAscii);
	UDPPixelFloodBin  .begin(pixelFloodPortBin);
	UDPBinFrames      .begin(binFramesPort);

	Serial.printf("Now listening at IP %s, UDP port %d pixelfloodAscii (bin: %d) and UDP port %d for binary frames\r\n", WiFi.localIP().toString().c_str(), pixelFloodPortAscii, pixelFloodPortBin, binFramesPort);

	last_ts = millis();

	printText(WiFi.localIP().toString().c_str());
}

char *strnchr(char *p, char what, int max_n) {
	char *const end = &p[max_n];

	while(p != end && *p != 0x00) {
		if (*p == what)
			return p;

		p++;
	}

	return nullptr;
}

int hex2toint(const char *p) {
	uint8_t v = 0;

	if (p[0] >= 'A')
		v = 10 + p[0] - 'A';
	else if (p[0] >= 'a')
		v = 10 + p[0] - 'a';
	else
		v = p[0] - '0';

	v <<= 4;

	if (p[1] >= 'A')
		v |= 10 + p[1] - 'A';
	else if (p[1] >= 'a')
		v |= 10 + p[1] - 'a';
	else
		v |= p[1] - '0';

	return v;
}

void pixelFloodAsciiProcessing(const unsigned int packetSize) {
	static uint32_t pts  = 0;
	static uint32_t npix = 0;
	static uint32_t npkt = 0;

	static char incomingPacket[1600];

	UDPPixelFloodAscii.read(incomingPacket, sizeof incomingPacket);
	incomingPacket[min(packetSize, sizeof(incomingPacket) - 1)] = 0x00;

	npkt++;

	char *p          = &incomingPacket[0];
	char *const pend = &incomingPacket[packetSize];

	while(pend - p >= 13) {
		char *lf = strnchr(p, '\n', pend - p);
		if (!lf)
			break;

		*lf = 0x00;

		if (p[0] == 'P' && p[1] == 'X') {
			p = &p[3];  // first digit after space

			int x = 0;
			while(p < lf && *p != ' ')
				x *= 10, x += *p - '0', p++;

			p++; // skip space

			int y = 0;
			while(p < lf && *p != ' ')
				y *= 10, y += *p - '0', p++;

			p++; // skip space

			if (x < 64 && y < 32 && lf - p >= 6) {
				int r = hex2toint(&p[0]);
				int g = hex2toint(&p[2]);
				int b = hex2toint(&p[4]);

				dma_display->drawPixelRGB888(x, y, r, g, b);

				npix++;
			}
		}

		p = lf + 1;
	}

	uint32_t fints = millis();
	uint32_t tdiff = fints - pts;

	if (tdiff >= 1000) {
		if (pts) {
			Serial.printf("PixelFloodAscii packets per second: %f\r\n", npkt * 1000. / tdiff);
			Serial.printf("PixelFloodAscii pixels per second: %f\r\n", npix * 1000. / tdiff);
		}

		pts = fints;

		npix = 0;
		npkt = 0;
	}
}

void pixelFloodBinProcessing(const unsigned int packetSize) {
	static uint32_t pts  = 0;
	static uint32_t npix = 0;
	static uint32_t npkt = 0;

	static char incomingPacket[1600];

	UDPPixelFloodBin.read(incomingPacket, sizeof incomingPacket);
	incomingPacket[min(packetSize, sizeof(incomingPacket) - 1)] = 0x00;

	npkt++;

	uint8_t *p          = reinterpret_cast<uint8_t *>(&incomingPacket[0]);
	uint8_t *const pend = reinterpret_cast<uint8_t *>(&incomingPacket[packetSize]);

	p += 2;  // header

	while(pend - p >= 7) {
		int x = (p[1] << 8) | p[0];
		int y = (p[3] << 8) | p[2];

		if (x < 64 && y < 32)
			dma_display->drawPixelRGB888(x, y, p[4], p[5], p[6]);

		p += 7;

		npix++;
	}

	uint32_t fints = millis();
	uint32_t tdiff = fints - pts;

	if (tdiff >= 1000) {
		if (pts) {
			Serial.printf("PixelFloodBin packets per second: %f\r\n", npkt * 1000. / tdiff);
			Serial.printf("PixelFloodBin pixels per second: %f\r\n", npix * 1000. / tdiff);
		}

		pts = fints;

		npix = 0;
		npkt = 0;
	}
}

void binFrameProcessing(const unsigned int packetSize) {
	static uint32_t pts  = 0;
	static uint32_t npix = 0;
	static uint32_t npkt = 0;

	static char incomingPacket[1600];

	UDPBinFrames.read(incomingPacket, sizeof incomingPacket);
	incomingPacket[min(packetSize, sizeof(incomingPacket) - 1)] = 0x00;

	npkt++;

	uint8_t *p          = reinterpret_cast<uint8_t *>(&incomingPacket[0]);
	uint8_t *const pend = reinterpret_cast<uint8_t *>(&incomingPacket[packetSize]);

	int x = 0;
	int y = 0;

	y += p[0] << 8;
	y += p[1];

	p += 2;

	while(pend - p >= 3) {
		dma_display->drawPixelRGB888(x, y, p[0], p[1], p[2]);

		p += 3;

		x++;

		if (x == 64) {
			x = 0;

			y++;

			if (y == 32)
				y = 0;
		}

		npix++;
	}

	uint32_t fints = millis();
	uint32_t tdiff = fints - pts;

	if (tdiff >= 1000) {
		if (pts) {
			Serial.printf("BinFrame packets per second: %f\r\n", npkt * 1000. / tdiff);
			Serial.printf("BinFrame pixels per second: %f\r\n", npix * 1000. / tdiff);
		}

		pts = fints;

		npix = 0;
		npkt = 0;
	}
}

void loop() {
	ArduinoOTA.handle();

	uint32_t now_ts = millis();

	if (!ota_enabled) {
		unsigned int pfAsciiPacketSize = UDPPixelFloodAscii.parsePacket();
		if (pfAsciiPacketSize) {
			last_ts = now_ts;
			pixelFloodAsciiProcessing(pfAsciiPacketSize);
		}

		unsigned int pfBinPacketSize = UDPPixelFloodBin.parsePacket();
		if (pfBinPacketSize) {
			last_ts = now_ts;
			pixelFloodBinProcessing(pfBinPacketSize);
		}

		unsigned int binPacketSize = UDPBinFrames.parsePacket();
		if (binPacketSize) {
			last_ts = now_ts;
			binFrameProcessing(binPacketSize);
		}
	}

	if (now_ts - last_ts >= 5000) {
		dma_display->clearScreen();

		last_ts = now_ts;
	}
}
